package com.schbrain.canal.client.transfer;

import com.alibaba.otter.canal.client.CanalConnector;
import com.alibaba.otter.canal.protocol.Message;
import com.alibaba.otter.canal.protocol.exception.CanalClientException;
import com.schbrain.canal.client.conf.CanalClientConfig;
import com.schbrain.canal.client.core.HandlerConf;
import lombok.extern.slf4j.Slf4j;
import java.util.Map;
import java.util.Objects;
import java.util.concurrent.atomic.AtomicInteger;

/**
 * @author zhuyf
 * @date 2022/6/16
 */
@Slf4j
public abstract class AbstractMessageTransponder implements MessageTransponder {

    /**
     * canal connector
     */
    private final CanalConnector connector;

    /**
     * custom config
     */
    protected final CanalClientConfig config;

    /**
     * destination of canal server
     */
    protected final String destination;

    protected final HandlerConf handlerConf;
    /**
     * running flag
     */
    private volatile boolean running = true;

    private AtomicInteger tryCount = new AtomicInteger(0);

    public AbstractMessageTransponder(CanalConnector connector,
                                      Map.Entry<String,CanalClientConfig> config,
                                      HandlerConf handlerConf) {
        Objects.requireNonNull(connector, "connector can not be null!");
        Objects.requireNonNull(config, "config can not be null!");
        this.connector = connector;
        this.destination = config.getKey();
        this.config = config.getValue();
        this.handlerConf = handlerConf;
    }

    @Override
    public void run() {
        tryCount.set(config.getRetryCount());
        final long interval = config.getAcquireInterval();
        while (running && !Thread.currentThread().isInterrupted()) {
            try {
                Message message = connector.getWithoutAck(config.getBatchSize());
                long batchId = message.getId();
                int size = message.getEntries().size();
                //empty message
                if (batchId == -1 || size == 0) {
                    Thread.sleep(interval);
                } else {
                    distributeEvent(message);
                }
                connector.ack(batchId);
            } catch (CanalClientException e) {
                boolean nextTry = nextTry();
                if(!nextTry){
                    stop();
                }
                try {
                    Thread.sleep(interval);
                } catch (InterruptedException e1) {
                }
            } catch (InterruptedException e) {
                stop();
                connector.rollback();
            } finally {
                if(!running){
                    connector.rollback();
                    break;
                }
            }
        }
        stop();
        try {
            connector.disconnect();
        }catch (Exception e){
            e.printStackTrace();
        }
        log.info("{}: client stopped. ", destination);
    }

    /**
     * 是否尝试下一次
     * @return
     */
    private boolean nextTry(){
        int retryCount = config.getRetryCount();
        if(retryCount <= 0){
            return true;
        }
        int n = tryCount.decrementAndGet();
        return n > 0;
    }

    /**
     * to distribute the message to special event and let the event listeners to handle it
     *
     * @param message canal message
     */
    protected abstract void distributeEvent(Message message);

    /**
     * stop running
     */
    void stop() {
        running = false;
    }

}
